bar_styles = { # [Foreground, Background]
    "normal": {
        "BAR": {
            "+": [None, None],
            "-": [None, None],
        },
        "DESC": [None, None],
        "CUR": [None, None],
        "TOTAL": [None, None],
        "PERCENT": [None, None],
        "SPEED": [None, None],
        "ETA": [None, None],
    },
    
    "simple": {
        "BAR": {
            "+": [0x7533F0, None],
            "-": [0x474747, None],
        },
        "DESC": [None, None],
        "CUR": [None, None],
        "TOTAL": [None, None],
        "PERCENT": [None, None],
        "SPEED": [None, None],
        "ETA": [None, None],
    },
    
    "contrast": {
        "BAR": {
            "+": [0xFFFFFF, 0x007ACC],
            "-": [0xCCCCCC, 0x1E1E1E],
        },
        "DESC": [0xFFD700, None],
        "CUR": [0x00FF00, None],
        "TOTAL": [0x00FFFF, None],
        "PERCENT": [0xFF00FF, None],
        "SPEED": [0xFFA500, None],
        "ETA": [0xFF4500, None],
    },

    "dark": {
        "BAR": {
            "+": [0x00FF00, 0x000000],
            "-": [0x444444, 0x000000],
        },
        "DESC": [0xAAAAAA, None],
        "CUR": [0x888888, None],
        "TOTAL": [0x666666, None],
        "PERCENT": [0x999999, None],
        "SPEED": [0x777777, None],
        "ETA": [0x555555, None],
    },

    "mono": {
        "BAR": {
            "+": [0xFFFFFF, None],
            "-": [0x888888, None],
        },
        "DESC": [0xAAAAAA, None],
        "CUR": [0xAAAAAA, None],
        "TOTAL": [0xAAAAAA, None],
        "PERCENT": [0xAAAAAA, None],
        "SPEED": [0xAAAAAA, None],
        "ETA": [0xAAAAAA, None],
    },

    "matrix": {
        "BAR": {
            "+": [0x00FF00, 0x000000], 
            "-": [0x003300, 0x000000], 
        },
        "DESC": [0x00FF00, None],
        "CUR": [0x00FF00, None],
        "TOTAL": [0x00FF00, None],
        "PERCENT": [0x00FF00, None],
        "SPEED": [0x009900, None],
        "ETA": [0x009900, None],
    },

    "solarized": {
        "BAR": {
            "+": [0x268BD2, 0x002B36],
            "-": [0x586E75, 0x002B36],
        },
        "DESC": [0xB58900, None],
        "CUR": [0x859900, None],
        "TOTAL": [0x2AA198, None],
        "PERCENT": [0xD33682, None],
        "SPEED": [0x6C71C4, None],
        "ETA": [0xCB4B16, None],
    },

    "gruvbox": {
        "BAR": {
            "+": [0xFABD2F, 0x282828], 
            "-": [0x3C3836, 0x282828],
        },
        "DESC": [0xFE8019, None],
        "CUR": [0xB8BB26, None],
        "TOTAL": [0xB8BB26, None],
        "PERCENT": [0xFABD2F, None],
        "SPEED": [0x8EC07C, None],
        "ETA": [0x83A598, None],
    },

    "nord": {
        "BAR": {
            "+": [0x88C0D0, 0x2E3440], 
            "-": [0x4C566A, 0x2E3440],
        },
        "DESC": [0x81A1C1, None],
        "CUR": [0xA3BE8C, None],
        "TOTAL": [0xA3BE8C, None],
        "PERCENT": [0x88C0D0, None],
        "SPEED": [0x5E81AC, None],
        "ETA": [0xBF616A, None],
    },

    "dracula": {
        "BAR": {
            "+": [0xBD93F9, 0x282A36], 
            "-": [0x44475A, 0x282A36],
        },
        "DESC": [0xFF79C6, None],
        "CUR": [0x50FA7B, None],
        "TOTAL": [0x50FA7B, None],
        "PERCENT": [0xBD93F9, None],
        "SPEED": [0x8BE9FD, None],
        "ETA": [0xFFB86C, None],
    },

    "firewatch": {
        "BAR": {
            "+": [0xFF6B35, 0x1E1E24],
            "-": [0x4A4E69, 0x1E1E24],
        },
        "DESC": [0xF7B32B, None],
        "CUR": [0x6FFFE9, None],
        "TOTAL": [0x6FFFE9, None],
        "PERCENT": [0xFF6B35, None],
        "SPEED": [0xA9BCD0, None],
        "ETA": [0xE36414, None],
    },

    "monochrome": {
        "BAR": {
            "+": [0xFFFFFF, 0x000000],  
            "-": [0x555555, 0x000000],
        },
        "DESC": [0xCCCCCC, None],
        "CUR": [0xCCCCCC, None],
        "TOTAL": [0xCCCCCC, None],
        "PERCENT": [0xFFFFFF, None],
        "SPEED": [0x888888, None],
        "ETA": [0x888888, None],
    },

    "pastel": {
        "BAR": {
            "+": [0xFFB3BA, 0xFAF3F3], 
            "-": [0xDADADA, 0xFAF3F3],
        },
        "DESC": [0xBAE1FF, None],
        "CUR": [0xFFDFBA, None],
        "TOTAL": [0xFFDFBA, None],
        "PERCENT": [0xB28DFF, None],
        "SPEED": [0xAFF8DB, None],
        "ETA": [0xFFFFBA, None],
    }
}

def hex_to_ansi(hex_color):
    if hex_color is None:
        return None, None
    if isinstance(hex_color, int):
        hex_color = f"{hex_color:06x}"
    if hex_color.startswith("0x"):
        hex_color = hex_color[2:]
    if len(hex_color) != 6:
        raise ValueError("Hex color must be 6 characters long (RGB)")
    r = int(hex_color[0:2], 16)
    g = int(hex_color[2:4], 16)
    b = int(hex_color[4:6], 16)
    code = 16 + 36 * (r // 43) + 6 * (g // 43) + (b // 43)
    return code, code

def style_text(text, fg_bg_pair):
    fg, bg = fg_bg_pair
    seq = ""
    if fg is not None:
        fg_code = hex_to_ansi(fg)[0]
        seq += f"\033[38;5;{fg_code}m"
    if bg is not None:
        bg_code = hex_to_ansi(bg)[1]
        seq += f"\033[48;5;{bg_code}m"
    return f"{seq}{text}\033[0m" if seq else text


def check_bar_style_syntax(bar_style):
    example_style = {
        "BAR": {
            "+": [0xFFFFFF, 0x000000],
            "-": [0x888888, None],
        },
        "DESC": [0xAAAAAA, None],
        "CUR": [0xAAAAAA, None],
        "TOTAL": [0xAAAAAA, None],
        "PERCENT": [0xAAAAAA, None],
        "SPEED": [0xAAAAAA, None],
        "ETA": [0xAAAAAA, None],
    }

    def err(msg):
        raise ValueError(f"{msg}\n\nExample style:\n{example_style}")

    if not isinstance(bar_style, dict):
        err("Bar style must be a dictionary.")

    if "BAR" not in bar_style or not isinstance(bar_style["BAR"], dict):
        err("Bar style must contain a 'BAR' key with a dictionary value.")

    for key in ["+", "-"]:
        if key not in bar_style["BAR"]:
            err(f"'BAR' must contain a '{key}' key.")
        if not isinstance(bar_style["BAR"][key], list):
            err(f"'BAR' key '{key}' must be a list.")

    required_keys = ["DESC", "CUR", "TOTAL", "PERCENT", "SPEED", "ETA"]
    for key in required_keys:
        if key not in bar_style:
            err(f"Bar style must contain a '{key}' key.")
        if not isinstance(bar_style[key], list):
            err(f"Bar style key '{key}' must be a list.")

    return True


import builtins
import threading
import time
import sys
import os
import requests
import aiohttp
import io
import shutil
from collections import deque
from .styles import bar_styles, style_text, check_bar_style_syntax

class SmartBar:
    _instances = []
    _io_patched = False
    _global_lock = threading.Lock()
    _position_map = {}

    def __init__(
        self,
        desc: str,
        length: int = 40,
        stream=sys.stdout,
        position: str = "/r",
        style: str = r"--",
        bar_style: str = "normal",
        auto_bar: bool = True,
        mode: str = "bytes",
        unit: str = None,
        custom_bar_output: str = None,
        recuv: bool = False,
        custom_bar_style: dict = None,
    ):
        self.desc = desc
        self.length = length
        self.stream = stream
        self.total = 0
        self.current = 0
        self.position = position
        self._start_time = None
        self._speed_window = deque(maxlen=5)
        self._running = False
        self._paused = False
        self._lock = threading.Lock()
        self._thread = None
        self._auto_total_set = False
        self.style = style
        self.auto_bar = auto_bar
        self.mode = mode
        self.unit = unit
        self.custom_bar_output = custom_bar_output
        self.recuv = recuv
        self.bar_styles = bar_styles
        self.bar_style = bar_style
        self.custom_bar_style = custom_bar_style

        if self.custom_bar_style and self.bar_style:
            raise ValueError("Cannot specify both 'custom_bar_style' and 'bar_style'.")
        if self.custom_bar_style:
            check_bar_style_syntax(self.custom_bar_style)
            self.bar_style = "custom"
            self.bar_styles["custom"] = self.custom_bar_style

        if self.bar_style not in self.bar_styles:
            raise ValueError(f"Invalid bar style: {self.bar_style}\nall Styles: {list(self.bar_styles.keys())}")
        
        self._bar_style = self.bar_styles[self.bar_style]
        

        if self.mode == "custom" and not self.unit:
            raise ValueError("Custom mode requires a 'unit' argument.")
        if self.position not in ["top", "bottom", "/r"]:
            raise ValueError("Position must be 'top', 'bottom', or '/r'.")
        if self.mode not in ["bytes", "steps", "items", "custom"]:
            raise ValueError("Mode must be 'bytes', 'steps', 'items', or 'custom'.")

    def __enter__(self):
        with SmartBar._global_lock:
            self._nested_level = len(SmartBar._instances)
            SmartBar._instances.append(self)
            self._assign_position()
            if self.auto_bar and not SmartBar._io_patched:
                self._patch_io()
                SmartBar._io_patched = True

        self._start_time = time.time()
        self._running = True
        self._paused = False
        self._thread = threading.Thread(target=self._refresh_loop, daemon=True)
        self._thread.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._running = False
        if self._thread:
            self._thread.join()

        with SmartBar._global_lock:
            if self in SmartBar._instances:
                SmartBar._instances.remove(self)
            SmartBar._position_map.pop(self, None)
            if self.auto_bar and not SmartBar._instances:
                self._restore_io()
                SmartBar._io_patched = False

        self._draw()
        self.stream.flush()

    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.__exit__(exc_type, exc_val, exc_tb)

    def _assign_position(self):
        lines = shutil.get_terminal_size().lines
        used_top = [SmartBar._position_map[b] for b in SmartBar._instances if b.position == "top"]
        used_bottom = [SmartBar._position_map[b] for b in SmartBar._instances if b.position == "bottom"]

        if self.position == "top":
            pos = 0
            while pos in used_top:
                pos += 1
            SmartBar._position_map[self] = pos
        elif self.position == "bottom":
            pos = lines - 1
            while pos in used_bottom:
                pos -= 1
            SmartBar._position_map[self] = pos
        else:
            SmartBar._position_map[self] = -1

    def _patch_io(self):
        self._orig_open = builtins.open
        builtins.open = self._patched_open

        self._orig_requests_get = requests.get
        requests.get = self._patched_requests_get

        self._orig_aiohttp_session = aiohttp.ClientSession._request
        aiohttp.ClientSession._request = self._patched_aiohttp_request

    def _restore_io(self):
        builtins.open = self._orig_open
        requests.get = self._orig_requests_get
        aiohttp.ClientSession._request = self._orig_aiohttp_session

    def _patched_open(self, *args, **kwargs):
        f = self._orig_open(*args, **kwargs)
        return PatchedFile(f)

    def _patched_requests_get(self, *args, **kwargs):
        resp = self._orig_requests_get(*args, **kwargs)
        try:
            transfer_encoding = resp.headers.get("Transfer-Encoding", "").lower()
            if "chunked" in transfer_encoding:
                self.total = 0
            else:
                length = int(resp.headers.get("Content-Length", 0))
                if length and not self._auto_total_set:
                    self.total = length
                    self._auto_total_set = True
        except:
            pass
        return PatchedResponse(resp)
    
    async def _patched_aiohttp_request(self, session, method, url, *args, **kwargs):
        resp = await self._orig_aiohttp_session(session, method, url, *args, **kwargs)
        try:
            transfer_encoding = resp.headers.get("Transfer-Encoding", "").lower()
            if "chunked" in transfer_encoding:
                self.total = 0
            else:
                length = int(resp.headers.get("Content-Length", 0))
                if length and not self._auto_total_set:
                    self.total = length
                    self._auto_total_set = True
        except:
            pass
        return PatchedAIOHTTPResponse(resp)

    def _refresh_loop(self):
        last_bytes = 0
        last_time = time.time()
        while self._running:
            if self._paused:
                time.sleep(0.1)
                continue

            now = time.time()
            elapsed = now - last_time
            if elapsed > 0.1:
                delta = self.current - last_bytes
                speed = delta / elapsed if elapsed > 0 else 0
                self._speed_window.append(speed)
                last_bytes = self.current
                last_time = now
            self._redraw_all()
            time.sleep(0.1)

    @classmethod
    def _redraw_all(cls):
        with cls._global_lock:
            sys.stdout.write("\0337")  
            sys.stdout.write("\033[?25l")  

            sorted_bars = sorted(cls._instances, key=lambda b: b._nested_level)

            for idx, bar in enumerate(sorted_bars):
                line = cls._position_map.get(bar, -1)
                if line >= 0:
                    sys.stdout.write(f"\033[{line + 1};0H")
                bar._draw(inline=(line == -1))

            sys.stdout.write("\0338")  
            sys.stdout.flush()

    def fmt_value(self, value):
        if self.mode == "bytes":
            for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
                if value < 1024:
                    return f"{value:.1f} {unit}"
                value /= 1024
            return f"{value:.1f} PB"
        elif self.mode in ("steps", "items", "custom"):
            return f"{value:.1f} {self.unit or ''}"
        return str(value)

    def _draw(self, inline=False):
        with self._lock:
            percent = self.current / self.total if self.total else 0
            percent = min(percent, 1.0)
            filled = int(self.length * percent)

            
            bar = ""
            bar_filled_char, bar_unfilled_char = self.style[0], self.style[1]
            bar_colors = self._bar_style["BAR"]

            fg_fill, bg_fill = bar_colors.get("+", [None, None])
            fg_unfill, bg_unfill = bar_colors.get("-", [None, None])

            for i in range(self.length):
                if i < filled:
                    bar += style_text(bar_filled_char, [fg_fill, bg_fill])
                else:
                    bar += style_text(bar_unfilled_char, [fg_unfill, bg_unfill])

            
            speed = sum(self._speed_window) / len(self._speed_window) if self._speed_window else 0
            window_speed = speed
            avg_speed = self.current / (time.time() - self._start_time + 1e-6)
            combined_speed = (0.7 * window_speed + 0.3 * avg_speed)
            eta = (self.total - self.current) / combined_speed if self.total and combined_speed else None

          
            cur_disp = self.fmt_value(self.current)
            total_disp = self.fmt_value(self.total) if self.total else "?"
            speed_disp = self.fmt_value(combined_speed) + "/s" if combined_speed else "-"
            eta_disp = f"{eta:.1f}s" if eta is not None and eta >= 0 else "?"

        
            DESC_C = self._bar_style["DESC"]
            CUR_C = self._bar_style["CUR"]
            TOTAL_C = self._bar_style["TOTAL"]
            PERCENT_C = self._bar_style["PERCENT"]
            SPEED_C = self._bar_style["SPEED"]
            ETA_C = self._bar_style["ETA"]

            desc_disp = style_text(self.desc, DESC_C)
            cur_disp = style_text(cur_disp, CUR_C)
            total_disp = style_text(total_disp, TOTAL_C)
            percent_disp = style_text(f"{percent * 100:.1f}%", PERCENT_C)
            speed_disp = style_text(speed_disp, SPEED_C)
            eta_disp = style_text(eta_disp, ETA_C)

            if self.custom_bar_output:
                output = self.custom_bar_output
                output = output.replace("%(DESC)", desc_disp)
                output = output.replace("%(BAR)", bar)
                output = output.replace("%(CUR)", cur_disp)
                output = output.replace("%(TOTAL)", total_disp)
                output = output.replace("%(PERCENT)", percent_disp)
                output = output.replace("%(SPEED)", speed_disp)
                output = output.replace("%(ETA)", eta_disp)
            else:
                output = (
                    f"{desc_disp:20} [{bar}] {cur_disp}/{total_disp} "
                    f"({percent_disp}) | {speed_disp} | ETA: {eta_disp}"
                )

            if inline:
                self.stream.write("\r\033[2K")
                self.stream.write(output)
                self.stream.write(" " * 10 + "\r")
            else:
                self.stream.write("\033[2K")
                self.stream.write(output + "\n")

            self.stream.flush()

    def pause(self):
        with self._lock:
            self._paused = True

    def resume(self):
        with self._lock:
            if not self._running:
                return
            self._paused = False

    def update(self, value):
        with self._lock:
            self.current = value
            if self.total and self.current > self.total:
                self.current = self.total

    def add(self, n):
        with self._lock:
            self.current += n
            if self.total and self.current > self.total:
                self.current = self.total

    @classmethod
    def ignore(cls, obj):
        if isinstance(obj, (PatchedFile, PatchedResponse, PatchedAIOHTTPResponse)):
            obj._ignore = True

    @classmethod
    def unignore(cls, obj):
        if isinstance(obj, (PatchedFile, PatchedResponse, PatchedAIOHTTPResponse)):
            obj._ignore = False
    
    def track(self, obj):
        if isinstance(obj, io.IOBase):
            orig_write = obj.write
            orig_read = obj.read

            def new_write(data):
                result = orig_write(data)
                self.add(len(data))
                return result

            def new_read(size=-1):
                data = orig_read(size)
                self.add(len(data))
                return data

            if hasattr(obj, 'write'):
                obj.write = new_write
            if hasattr(obj, 'read'):
                obj.read = new_read
            return obj


class PatchedFile:
    def __init__(self, file_obj):
        self._file = file_obj
        self._ignore = False
        self._init_size()

    def _init_size(self):
        try:
            instance = SmartBar._instances[-1]
            if instance and not instance._auto_total_set:
                if hasattr(self._file, 'name') and os.path.isfile(self._file.name):
                    size = os.path.getsize(self._file.name)
                    if size > 0:
                        instance.total = size
                        instance._auto_total_set = True
        except:
            pass

    def read(self, size=-1):
        data = self._file.read(size)
        if not self._ignore:
            SmartBar._instances[-1].add(len(data))
        return data

    def write(self, data):
        if not self._ignore:
            SmartBar._instances[-1].add(len(data))
        return self._file.write(data)

    def __getattr__(self, name):
        return getattr(self._file, name)

    def __enter__(self):
        self._file.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._file.__exit__(exc_type, exc_val, exc_tb)


class PatchedResponse:
    def __init__(self, resp, chunk_size=8192):
        self._resp = resp
        self._chunk_size = chunk_size
        self._ignore = False

    def iter_content(self, chunk_size=None):
        for chunk in self._resp.iter_content(chunk_size or self._chunk_size):
            if not self._ignore:
                SmartBar._instances[-1].add(len(chunk))
            yield chunk

    def __getattr__(self, name):
        return getattr(self._resp, name)


class PatchedAIOHTTPResponse:
    def __init__(self, resp):
        self._resp = resp
        self._ignore = False

    async def read(self):
        data = await self._resp.read()
        if not self._ignore:
            SmartBar._instances[-1].add(len(data))
        return data

    async def iter_chunked(self, chunk_size=8192):
        async for chunk in self._resp.content.iter_chunked(chunk_size):
            if not self._ignore:
                SmartBar._instances[-1].add(len(chunk))
            yield chunk

    def __getattr__(self, name):
        return getattr(self._resp, name)



from setuptools import setup, find_packages
import re; import os


with open(os.path.join("README.md"), "r", encoding="utf-8") as file:
    long_description = file.read()

def get_version():
    with open(os.path.join("smartbar", "__init__.py")) as f:
        content = f.read()
    match = re.search(r'__version__ = ["\']([^"\']+)["\']', content)
    return match.group(1)


setup(
    name='smartbar',
    version=get_version(),
    description='Smartbar',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='lama2923',
    author_email='lama2923.v2@gmail.com',
    project_urls={
        'Author GitHub': 'https://github.com/lama2923'
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    keywords='example project development bar progress progressbar smartbar lama2923 design art custom custombar',
    packages=find_packages(),
    install_requires=[
        'requests',
        "aiohttp"
    ],
    python_requires='>=3.7',
)

